import os
import re
import typing
from pathlib import Path

import pandas as pd

from mlservicewrapper.core import errors


class NameValidator:
    __valid_name_regex = re.compile(r'^[A-Z][a-zA-Z\d]*$')

    @classmethod
    def raise_if_invalid(cls, name: str):
        if not cls.is_valid(name):
            raise ValueError("Name is not valid: '{}'!".format(name))
    
    @classmethod
    def is_valid(cls, name: str):
        return cls.__valid_name_regex.match(name) is not None
    
__all__ = ["NameValidator", "ServiceContext", "ProcessContext", "CollectingProcessContext", "EnvironmentVariableServiceContext", "CoalescingServiceContext"]

class ServiceContext:
    def get_parameter_value(self, name: str, required: bool = True, default: str = None) -> str:
        raise NotImplementedError()

    def get_parameter_real_path_value(self, name: str, required: bool = True) -> Path:
        NameValidator.raise_if_invalid(name)

        val = self.get_parameter_value(name, required)

        if val is None:
            return val

        return os.path.join(os.getcwd(), val)
        
class ProcessContext:
    def get_parameter_value(self, name: str, required: bool = True, default: str = None) -> str:
        raise NotImplementedError()
    
    async def get_input_dataframe(self, name: str, required: bool = True) -> pd.DataFrame:
        raise NotImplementedError()

    async def set_output_dataframe(self, name: str, df: pd.DataFrame):
        raise NotImplementedError()
        
class CollectingProcessContext(ProcessContext):
    def __init__(self):
        super().__init__()
        self.__output_dataframes: typing.Dict[str, pd.DataFrame] = dict()

    async def set_output_dataframe(self, name: str, df: pd.DataFrame):
        NameValidator.raise_if_invalid(name)

        self.__output_dataframes[name] = df
    
    def get_output_dataframe(self, name: str) -> pd.DataFrame:
        NameValidator.raise_if_invalid(name)

        return self.__output_dataframes.get(name)

    def output_dataframes(self):
        return self.__output_dataframes.items()

class CoalescingServiceContext(ServiceContext):
    def __init__(self, contexts: typing.List[ServiceContext]):
        super().__init__()
        self.__contexts = contexts
    
    def get_parameter_value(self, name: str, required: bool = True, default: str = None) -> str:
        NameValidator.raise_if_invalid(name)

        for context in self.__contexts:
            val = context.get_parameter_value(name, False)

            if val is not None:
                return val
            
        if required and default is None:
            raise errors.MissingParameterError(name)

        return default

    def get_parameter_real_path_value(self, name: str, required: bool = True) -> Path or None:
        NameValidator.raise_if_invalid(name)

        for context in self.__contexts:
            val = context.get_parameter_real_path_value(name, False)

            if val is not None:
                return val
            
        if required:
            raise errors.MissingParameterError(name)

        return None

class DictServiceContext(ServiceContext):
    def __init__(self, values: dict):
        super().__init__()
        self.__values = values
    
    def get_parameter_value(self, name: str, required: bool = True, default: str = None) -> str:
        NameValidator.raise_if_invalid(name)

        ev = self.__values.get(name) or default

        if required and not ev:
            raise errors.MissingParameterError(name)

        return ev


class EnvironmentVariableServiceContext(ServiceContext):
    def __init__(self, prefix: str):
        super().__init__()
        self.__prefix = prefix
    
    def get_parameter_value(self, name: str, required: bool = True, default: str = None) -> str:
        NameValidator.raise_if_invalid(name)

        ev = os.environ.get(self.__prefix + name) or default

        if required and not ev:
            raise errors.MissingParameterError(name)

        return ev



import json
import os
import typing
from pathlib import Path

def _get_real_path(base_path: str, relative: str) -> Path:
    if base_path is None:
        base_path_dir = os.getcwd()
    else:
        base_path_dir = os.path.dirname(base_path)
    
    return Path(os.path.realpath(os.path.join(base_path_dir, relative)))
    
class ConfigurationFile:
    def __init__(self, path: str = None, config: dict = None):
        if path is None:
            self.__config = config.copy()
        else:
            self.__path = path
                
            with open(path, "r") as config_file:
                self.__config = json.load(config_file)
            
        extends_path = self.__config.get("extends")

        if extends_path is None:
            self.__extends = None
        else:
            if isinstance(extends_path, str):
                extends_path = _get_real_path(self.__path, extends_path)

                self.__extends = ConfigurationFile(path=extends_path)
            else:
                raise ValueError("Invalid configuration file! The 'extends' property, if present, must be a string.")


    def _get_real_path(self, path: str) -> Path:
        return _get_real_path(self.__path, path)
        
    def has_value(self, name: typing.Union[str, typing.List[str]]) -> bool:
        return self.get_value(name) is not None
        
    def get_value(self, name: typing.Union[str, typing.List[str]]) -> typing.Any:
        val, _ = self.__get_value_with_source(name)

        return val

    def get_real_path_value(self, name: typing.Union[str, typing.List[str]]) -> Path:
        val, s = self.__get_value_with_source(name)

        return s._get_real_path(val)
        
    def __get_value_with_source(self, name: typing.Union[str, typing.List[str]]):
        if isinstance(name, str):
            result = self.__config.get(name)
        else:
            if name is None or len(name) == 0:
                raise ValueError("Provide at least one name part")
            
            result = self.__config
            for part in name:
                result = result.get(part)

                if result is None:
                    break
        
        if result is None and self.__extends is not None:
            return self.__extends.__get_value_with_source(name)

        return result, self

class ParameterSchema:
    def __init__(self, name: str, d: dict) -> None:
        self.name = name
        self.required = d.get("required", True)
        self.type = d["type"]

class ParametersSchema:
    def __init__(self, d: typing.Dict[str, typing.Any]) -> None:
        self._data = d

    def items(self) -> typing.Iterable[ParameterSchema]:
        return (ParameterSchema(k, v) for k, v in self._data.items())

class DatasetSchema:
    def __init__(self, name: str, d: dict) -> None:
        self.name = name
        self.required = d.get("required", True)
        self.dataset_type = d.get("datasetType")
        self.item_schema = d.get("itemSchema")

class ServiceSchema:
    def __init__(self, schema_spec: dict) -> None:
        self._schema_spec = schema_spec

    def _get_parameters_schema(self, name: str):
        p: typing.Dict[str, dict] = self._schema_spec.get("parameters", dict())

        ret = p.get(name, dict())

        return ParametersSchema(ret)

    def _get_dataset_specs(self, direction: str) -> typing.Iterable[DatasetSchema]:
        if self._schema_spec is None:
            return list()

        datasets: typing.Dict[str, typing.Dict[str, dict]] = self._schema_spec.get("datasets")

        if datasets is None:
            return list()

        dataset_dict = datasets.get(direction)

        if dataset_dict is None:
            return list()

        return (DatasetSchema(k, v) for k, v in dataset_dict.items())
        
    def parameters_load(self):
        return self._get_parameters_schema("load")
        
    def parameters_process(self):
        return self._get_parameters_schema("process")

    def datasets_input(self):
        return self._get_dataset_specs("input")
        
    def datasets_output(self):
        return self._get_dataset_specs("output")

class ServiceInfo:
    def __init__(self, vals: dict) -> None:
        self.name = vals.get("name")
        self.version = vals.get("version")

class ServiceHostParameters:
    def __init__(self, vals: dict) -> None:
        self._vals = vals

    def get_host_config(self, name: str):
        d: dict = self._vals.get(name)

        if not d:
            return dict()

        return d

class ServiceParameters:
    def __init__(self, config: ConfigurationFile) -> None:
        self._config = config

    def get_value(self, name: str):
        return self._config.get_value(["parameters", name])

    def get_real_path_value(self, name: str):
        return self._config.get_real_path_value(["parameters", name])

class ServiceLoadParameters:
    def __init__(self, config: ConfigurationFile) -> None:
        self.module_path = config.get_real_path_value("modulePath")
        self.module_name = config.get_value("moduleName")
        self.package_name = config.get_value("packageName")

        if self.module_name and self.module_path:
            raise ValueError("Both modulePath and moduleName were specified!")
        elif not self.module_name and not self.module_path:
            raise ValueError("The module couldn't be determined! Please include either a modulePath or moduleName in configuration.")

        self.class_name = config.get_value("className")
        self.service_instance_name = config.get_value("serviceInstanceName")

        if self.class_name and self.service_instance_name:
            raise ValueError("Both className and serviceInstanceName were specified!")
        elif not self.class_name and not self.service_instance_name:
            raise ValueError("Either className or serviceInstanceName must be specified!")


T = typing.TypeVar("T")
class ServiceConfiguration:
    def __init__(self, config_file: typing.Union[str, ConfigurationFile]) -> None:
        if isinstance(config_file, str):
            config_file = ConfigurationFile(config_file)
        
        self._config = config_file

    def _has(self, name: str):
        d = self._config.get_value(name)

        return d is not None

    def _get_typed(self, type: typing.Callable[[dict], T], name: str, coalesce = False) -> T:
        d = self._config.get_value(name)

        if not d:
            if coalesce:
                return type(dict())
            else:
                return None

        return type(d)

    def service(self):
        return ServiceLoadParameters(self._config)

    def schema(self):
        return self._get_typed(ServiceSchema, "schema", coalesce=True)

    def info(self):
        return self._get_typed(ServiceInfo, "info", coalesce=True)

    def host(self):
        return self._get_typed(ServiceHostParameters, "host", coalesce=True)

    def has_parameters(self):
        return self._has("parameters")

    def parameters(self):
        return ServiceParameters(self._config)



import os

import mlservicewrapper
import mlservicewrapper.core.debug
from service import SampleService

root_dir = os.path.join(os.path.dirname(__file__), '..')

input_data_dir = os.path.join(root_dir, 'data', 'input')
output_data_dir = os.path.join(root_dir, 'data', 'output')

load_parameters = {
    "ModBy": "3"
}

split_dataset_name = "Data"

if __name__ == "__main__":

    results = mlservicewrapper.core.debug.run(
        SampleService,
        input_data_dir,
        split_dataset_name=split_dataset_name,
        load_parameters=load_parameters,
        output_file_directory=output_data_dir
    )

    predictions = results["Results"]




import typing

import os

import pandas as pd

class BaseFileHandler:
    def __init__(self, path: os.PathLike) -> None:
        self.path = path

    def read_dataframe(self) -> pd.DataFrame:
        raise NotImplementedError()

    def save_dataframe(self, df: pd.DataFrame) -> None:
        raise NotImplementedError()

class _CsvFileHandler(BaseFileHandler):
    def __init__(self, path: os.PathLike) -> None:
        super().__init__(path)

    def read_dataframe(self) -> pd.DataFrame:
        return pd.read_csv(self.path, keep_default_na=False)

    def save_dataframe(self, df: pd.DataFrame) -> None:
        return df.to_csv(self.path, index=False)
        
class _ExcelFileHandler(BaseFileHandler):
    def __init__(self, path: os.PathLike) -> None:
        super().__init__(path)

    def read_dataframe(self) -> pd.DataFrame:
        return pd.read_excel(self.path, keep_default_na=False)

    def save_dataframe(self, df: pd.DataFrame) -> None:
        return df.to_excel(self.path, index=False)

class FileDatasetLookup:
    def __init__(self, directory: str, path_map: typing.Dict[str, str]):
        self.__directory = directory
        self.__map = path_map

        self.__handler_maps: typing.Dict[os.PathLike, typing.Callable[[str], BaseFileHandler]] = {
            "csv": _CsvFileHandler,
            "xlsx": _ExcelFileHandler
        }

    def allowed_file_extensions(self) -> typing.Iterable[str]:
        return self.__handler_maps.keys()

    def _get_file_handler_from_path(self, path: os.PathLike) -> BaseFileHandler:
        _, ext = os.path.splitext(path)

        handler = self.__handler_maps.get(ext[1:])

        if not handler:
            raise NotImplementedError(f"File extension '{ext}' is not known!")
            
        return handler(path)

    def get_file_handler(self, name: str, default_extension: str or None = None, must_exist = True):
        if self.__map is not None and name in self.__map:
            return self._get_file_handler_from_path(self.__map[name])

        if self.__directory is None:
            return None

        if default_extension:
            if default_extension not in self.__handler_maps:
                raise ValueError("Default extension '{}' is not known!".format(default_extension))

            exts = [default_extension, *self.allowed_file_extensions()]
        else:
            exts = self.allowed_file_extensions()

        for ext in exts:
            path = os.path.join(self.__directory, name + os.path.extsep + ext)

            if path is None:
                continue

            if must_exist and not os.path.exists(path):
                continue

            return self._get_file_handler_from_path(path)

        return None



import argparse
import logging
import os

from .. import server
from . import local


#https://stackoverflow.com/a/42355279/1270504
class _StoreDictKeyPair(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        my_dict = {}
        for kv in values.split(","):
            k,v = kv.split("=")
            my_dict[k] = v
        setattr(namespace, self.dest, my_dict)

parser = argparse.ArgumentParser(description='Locally debug.', prog = "mlservicewrapper.core.debug")
parser.add_argument(
    '--config', help='Path to service configuration file', required=True)

parser.add_argument('--input-dir', dest='input_dir', help='Path to input directory')
parser.add_argument('--input-paths', dest='input_paths', help='Map input datasets to paths', action=_StoreDictKeyPair, metavar="<dataset name 1>=<path>,<dataset name 2>=<path>...")

parser.add_argument('--output-dir', dest='output_dir', help='Path to input directory')
parser.add_argument('--output-paths', dest='output_paths', help='Map output datasets to paths', action=_StoreDictKeyPair, metavar="<dataset name 1>=<path>,<dataset name 2>=<path>...")

parser.add_argument("--load-params", dest="load_params", action=_StoreDictKeyPair, metavar="KEY1=VAL1,KEY2=VAL2...")
parser.add_argument("--run-params", dest="runtime_parameters", action=_StoreDictKeyPair, metavar="KEY1=VAL1,KEY2=VAL2...")

parser.add_argument("--log-level")

parser.add_argument('--split-dataset-for-perf', dest='split_dataset_name',
                    help='Input dataset to split for performance evaluation.')

parser.add_argument('--assess-accuracy', dest='assess_accuracy', action=_StoreDictKeyPair, metavar="INPUT_DATASET_1.FIELD=OUTPUT_DATASET_1.FIELD,INPUT_DATASET_2.FIELD=OUTPUT_DATASET_2.FIELD,...",
                    help='Perform accuracy assessment against the given fields.')

parser.add_argument('--profile-processing-to-file', dest='profile_processing_to_file',
                    help='Enables profiling for the process phase, and writes results to a file which can then be read with pstats.')


args = parser.parse_args()

if args.log_level is not None:
    logging.basicConfig(level=args.log_level)

instance = server.ServerInstance(args.config)

local.run(
    instance,

    load_context=instance.build_context(override=args.load_params),

    runtime_parameters=args.runtime_parameters,

    input_dataset_paths=args.input_paths,
    input_dataset_directory=args.input_dir,

    split_dataset_name=args.split_dataset_name,
    
    output_dataset_directory=args.output_dir,
    output_dataset_paths=args.output_paths,
    
    assess_accuracy=args.assess_accuracy,

    profile_processing_to_file=args.profile_processing_to_file
)

instance.dispose()



import importlib

import inspect
import json
import os
import sys
import typing
from pathlib import Path

from mlservicewrapper.core.internal.service_loading import get_service_loader
import mlservicewrapper.core.configuration as configuration

from mlservicewrapper.core import contexts, errors, services

__all__ = ["ServerInstance"]

class _ConfigurationFileServiceContext(contexts.ServiceContext):
    def __init__(self, config: configuration.ServiceConfiguration):
        self.__parameters = config.parameters()
    
    def get_parameter_value(self, name: str, required: bool = True, default: str = None) -> str:
        contexts.NameValidator.raise_if_invalid(name)

        val = self.__parameters.get_value(name) or default
        
        if required and val is None:
            raise errors.MissingParameterError(name)

        return val

    def get_parameter_real_path_value(self, name: str, required: bool = True) -> Path:
        contexts.NameValidator.raise_if_invalid(name)

        val = self.__parameters.get_real_path_value(name)
        
        if required and val is None:
            raise errors.MissingParameterError(name)

        return val


class ServerInstance:
    def __init__(self, config_path: str = None):

        if not config_path:
            config_path = os.environ.get("SERVICE_CONFIG_PATH", "./service/config.json")

        self.__config = configuration.ServiceConfiguration(config_path)

        self.__service_instance_loader = get_service_loader(self.__config.service())
        
        self.__host_configs = self.__config.host()

        self.__schema: configuration.ServiceSchema = self.__config.schema()
        self.__info: configuration.ServiceInfo = self.__config.info()
        
        self.__service: services.Service = None

    def get_info(self) -> configuration.ServiceInfo:
        return self.__info

    def get_load_parameter_specs(self) -> configuration.ParametersSchema:
        return self.__schema.parameters_load()
        
    def get_process_parameter_specs(self) -> configuration.ParametersSchema:
        return self.__schema.parameters_process()
        
    def get_input_dataset_specs(self) -> typing.Iterable[configuration.DatasetSchema]:
        return self.__schema.datasets_input()

    def get_output_dataset_specs(self) -> typing.Iterable[configuration.DatasetSchema]:
        return self.__schema.datasets_output()

    def get_host_config_section(self, name: str) -> dict or None:
        if self.__host_configs is None:
            return None
        
        return self.__host_configs.get_host_config(name)

    def get_parameter_real_path_value(self, name: str) -> os.PathLike:
        return self.__config.parameters().get_real_path_value(name)
    
    def __get_service_instance(self) -> services.Service:

        service = self.__service_instance_loader.get_instance()

        print("Got service: {}".format(service))

        return service

    def is_loaded(self):
        return self.__service is not None

    def build_context(self, include_environment_variables = True, override: dict = None):
        context_parts: typing.List[contexts.ServiceContext] = []
        
        if override is not None:
            context_parts.append(contexts.DictServiceContext(override))

        if include_environment_variables:
            context_parts.append(contexts.EnvironmentVariableServiceContext("SERVICE_"))
        
        if self.__config.has_parameters():
            context_parts.append(_ConfigurationFileServiceContext(self.__config))

        return contexts.CoalescingServiceContext(context_parts)


    async def load(self, ctx: contexts.ServiceContext = None):
        service = self.__get_service_instance()
        
        if hasattr(service, 'load'):
            if ctx is None:
                ctx = self.build_context()

            print("service.load")
            load_result = service.load(ctx)

            if inspect.iscoroutine(load_result):
                await load_result

        self.__service = service

    async def process(self, ctx: contexts.ProcessContext):
        if not self.is_loaded():
            raise ValueError("Be sure to call load before process!")
        
        process_result = self.__service.process(ctx)

        if inspect.iscoroutine(process_result):
            await process_result

    def dispose(self):
        if self.__service is None or not hasattr(self.__service, 'dispose'):
            return
        
        self.__service.dispose()



import asyncio

import pandas as pd

import mlservicewrapper
import mlservicewrapper.core.contexts
import mlservicewrapper.core.services
import mlservicewrapper.core.errors

class SampleService(mlservicewrapper.core.services.Service):
    async def load(self, ctx: mlservicewrapper.core.contexts.ServiceContext):
        self.__mod_by = int(ctx.get_parameter_value("ModBy", default="2"))

    async def process(self, ctx: mlservicewrapper.core.contexts.ProcessContext):
        input_data = await ctx.get_input_dataframe("Data")

        if "TextField" not in input_data.columns:
            raise mlservicewrapper.core.errors.MissingDatasetFieldError("Data", "TextField")

        result = pd.DataFrame(input_data["TextField"].str.len() % self.__mod_by)

        result.columns = ["Result"]
        result.insert(0, "Id", input_data["Id"])

        await ctx.set_output_dataframe("Results", result)




from mlservicewrapper.core.contexts import ProcessContext, ServiceContext

__all__ = ["Service"]

class Service:
    async def load(self, ctx: ServiceContext):
        """Initialize variables and load models."""

    async def process(self, ctx: ProcessContext):
        """Run a prediction or processing service."""
        """Implementations may make in-place modifications to any data they receive."""

        raise NotImplementedError()

    def dispose(self):
        """Clean up any resources (file handles, temporary files, etc.) to gracefully shut down."""

        pass




import os

import importlib

from mlservicewrapper.core.configuration import ConfigurationFile, ServiceConfiguration, ServiceLoadParameters
import mlservicewrapper.core.services as services

class _BaseServiceModuleLoader:
    def load(self):
        raise NotImplementedError()

class _PathServiceModuleLoader(_BaseServiceModuleLoader):
    def __init__(self, path: str) -> None:
        super().__init__()

        self.__path = path

    def load(self):
        
        print("Loading from script {}".format(self.__path))

        dirname = os.path.dirname(self.__path)
        basename = os.path.basename(self.__path)

        os.sys.path.insert(0, dirname)

        module_name = os.path.splitext(basename)[0]

        print("Importing module {} from {}...".format(module_name, dirname))

        return importlib.import_module(module_name)

class _PackageServiceModuleLoader(_BaseServiceModuleLoader):
    def __init__(self, module_name: str, package_name: str) -> None:
        super().__init__()

        self.__module_name = module_name
        self.__package_name = package_name

    def load(self):
        if self.__package_name:
            print("Importing module {} from {}...".format(self.__module_name, self.__package_name))
        else:
            print("Importing module {}...".format(self.__module_name))

        return importlib.import_module(self.__module_name, self.__package_name)

def _get_service_module_loader(config: ServiceLoadParameters) -> _BaseServiceModuleLoader:
    if config.module_path:
        return _PathServiceModuleLoader(config.module_path)

    elif config.module_name:
        return _PackageServiceModuleLoader(config.module_name, config.package_name)

    else:
        raise ValueError("The module couldn't be determined! Please include either a modulePath or moduleName in configuration.")

class BaseServiceLoader:
    def __init__(self, module_loader: _BaseServiceModuleLoader) -> None:
        self.__module_loader = module_loader
    
    def _get_instance_from_module(self, service_module) -> services.Service:
        raise NotImplementedError()
    
    def get_instance(self):
        module = self.__module_loader.load()
        
        print("Imported module")

        return self._get_instance_from_module(module)

class _ClassServiceLoader(BaseServiceLoader):
    def __init__(self, module_loader: _BaseServiceModuleLoader, class_name: str) -> None:
        super().__init__(module_loader)

        self._class_name = class_name

    def _get_instance_from_module(self, service_module) -> services.Service:
        service_type = getattr(service_module, self._class_name)

        print("Identified service type: {}".format(str(service_type)))

        return service_type()

class _InstanceServiceLoader(BaseServiceLoader):
    def __init__(self, module_loader: _BaseServiceModuleLoader, instance_name: str) -> None:
        super().__init__(module_loader)

        self._instance_name = instance_name

    def _get_instance_from_module(self, service_module) -> services.Service:

        return getattr(service_module, self._instance_name)


def get_service_loader(config: ServiceLoadParameters) -> BaseServiceLoader:

    module_loader = _get_service_module_loader(config)
    
    if config.class_name:
        return _ClassServiceLoader(module_loader, config.class_name)

    elif config.service_instance_name:
        return _InstanceServiceLoader(module_loader, config.service_instance_name)
    else:
        raise ValueError("Either className or serviceInstanceName must be specified!")




class BaseProfilerWrapper:
    def enable(self) -> None:
        raise NotImplementedError()
    def disable(self) -> None:
        raise NotImplementedError()
    def save(self) -> None:
        raise NotImplementedError()

class _ProfilerWrapper(BaseProfilerWrapper):
    def __init__(self, to_path: str) -> None:
        super().__init__()
        
        self._to_path = to_path
        
        import cProfile
        self._profiler = cProfile.Profile()
    
    def enable(self):
        return self._profiler.enable()

    def disable(self):
        return self._profiler.disable()

    def save(self):
        return self._profiler.dump_stats(self._to_path)

class _EmptyProfilerWrapper(BaseProfilerWrapper):
    def enable(self):
        return None

    def disable(self):
        return None

    def save(self):
        return None

def get_profiler(profile_processing_to_file: str) -> BaseProfilerWrapper:
    
    if profile_processing_to_file is None:
        return _EmptyProfilerWrapper()
    
    return _ProfilerWrapper(profile_processing_to_file)



import asyncio
import json
import logging
import math
import os
import re
import statistics
import time
import typing

import pandas as pd

from mlservicewrapper.core.debug.file_lookups import FileDatasetLookup
from mlservicewrapper.core.debug.profiling import BaseProfilerWrapper, get_profiler

from .. import contexts, errors, services


def _print_ascii_histogram(seq: typing.List[float]) -> None:
    """A horizontal frequency-table/histogram plot."""

    hist = {}

    _min = min(seq)
    _max = max(seq)
    _len = len(seq)

    buckets = 10
    step = (_max - _min) / (buckets - 1)

    for i in seq:
        e = _min + (math.floor((i - _min) / step) * step)

        hist[e] = hist.get(e, 0) + 1

    for i in range(buckets):
        e = _min + (i * step)

        ct = hist.get(e, 0)

        pct = ct / _len

        w = math.floor(40 * pct)

        if ct > 0:
            w = max(w, 1)

        print('{0:5f}s {1}'.format(e, '+' * w))

def get_input_dataframe(name: str, file_lookup: FileDatasetLookup) -> pd.DataFrame:
    contexts.NameValidator.raise_if_invalid(name)

    file_handler = file_lookup.get_file_handler(name)

    return file_handler.read_dataframe()

class _LocalRunContext(contexts.CollectingProcessContext):
    def __init__(self, input_datasets: FileDatasetLookup, output_datasets: FileDatasetLookup = None, parameters: dict = None):
        super().__init__()
        self.__parameters = parameters or dict()

        self.input_datasets = input_datasets
        self.output_datasets = output_datasets

    def get_parameter_value(self, name: str, required: bool = True, default: str = None) -> str:
        contexts.NameValidator.raise_if_invalid(name)
        
        if name in self.__parameters:
            return self.__parameters[name]
        
        if required and default is None:
            raise errors.MissingParameterError(name)

        return default

    async def get_input_dataframe(self, name: str, required: bool = True):
        df = get_input_dataframe(name, self.input_datasets)
        
        if required and df is None:
            raise errors.MissingDatasetError(name)

        return df

    async def set_output_dataframe(self, name: str, df: pd.DataFrame):
        contexts.NameValidator.raise_if_invalid(name)

        await super().set_output_dataframe(name, df)

        handler = self.output_datasets.get_file_handler(name, "csv", must_exist=False)
        
        if not handler:
            return

        d = os.path.dirname(handler.path)
        os.makedirs(d, exist_ok=True)

        handler.save_dataframe(df)

class _LocalDataFrameRunContext(contexts.ProcessContext):
    def __init__(self, df: pd.DataFrame, name: str, base_ctx: contexts.ProcessContext):
        self.__base_ctx = base_ctx
        self.__name = name
        self.__df = df

    def get_parameter_value(self, name: str, required: bool = True, default: str = None) -> str:
        return self.__base_ctx.get_parameter_value(name, required, default)

    def set_output_dataframe(self, name: str, df: pd.DataFrame):
        return self.__base_ctx.set_output_dataframe(name, df)

    async def get_input_dataframe(self, name: str, required: bool = True):
        contexts.NameValidator.raise_if_invalid(name)
        
        if name == self.__name:
            return self.__df

        return await self.__base_ctx.get_input_dataframe(name, required)

async def _perform_accuracy_assessment(ctx: contexts.CollectingProcessContext, specs: typing.Dict[str, str]):

    for k, v in specs.items():
        i = k.split(".")
        o = v.split(".")

        input_df = await ctx.get_input_dataframe(i[0], required=True)
        output_df = ctx.get_output_dataframe(o[0])

        input_field = input_df[i[1]]
        input_field.name = "Expected"

        output_field = output_df[o[1]]
        output_field.name = "Actual"

        joined = output_field.to_frame().join(input_field, how="inner")

        joined["Result"] = joined["Actual"] == joined["Expected"]

        count_total = len(joined.index)
        count_correct = joined["Result"].values.sum()

        print("Accuracy ({} to {}): {} of {} ({})".format(k, v, count_correct, count_total, count_correct / count_total))

class _ServiceProcessTimer:
    def __init__(self, service: services.Service, profiler: "BaseProfilerWrapper") -> None:
        self._times: typing.List[float] = list()

        self._service = service
        self._profiler = profiler

    async def process(self, ctx: contexts.ProcessContext):
        
        s = time.perf_counter()
        self._profiler.enable()
        await self._service.process(ctx)
        self._profiler.disable()
        e = time.perf_counter()

        self._times.append(e - s)

    def print_stats(self):
        if len(self._times) == 0:
            print("Count: 0")
        elif len(self._times) == 1:
            print("Process time: {}s".format(self._times[0]))
        else:
            print()
            print("Count: {}".format(len(self._times)))
            print("Min process time: {}s".format(min(self._times)))
            print("Mean process time: {}s".format(statistics.mean(self._times)))
            print("Median process time: {}s".format(statistics.median(self._times)))
            print("Max process time: {}s".format(max(self._times)))

            _print_ascii_histogram(self._times)

def _get_run_contexts(run_context: _LocalRunContext, split_dataset_name: str):
    df = get_input_dataframe(split_dataset_name, run_context.input_datasets)
    
    for _, r in df.iterrows():
        rdf = pd.DataFrame([r]).reset_index(drop=True)
    
        row_run_context = _LocalDataFrameRunContext(rdf, split_dataset_name, run_context)

        yield row_run_context

async def _call_load(service: services.Service, load_context: contexts.ServiceContext = None):
    if load_context is None:
        load_context = contexts.DictServiceContext(dict())

    if hasattr(service, 'load'):
        print("Loading...")
        s = time.perf_counter()
        await service.load(load_context)
        e = time.perf_counter()

        load_time = e - s
    else:
        load_time = 0
    
    return load_time

def _get_run_context(input_dataset_paths: typing.Dict[str, str] = None, input_dataset_directory: str = None, output_dataset_directory: str = None, output_dataset_paths: typing.Dict[str, str] = None, runtime_parameters: dict = None):
    
    input_datasets = FileDatasetLookup(input_dataset_directory, input_dataset_paths)
    output_datasets = FileDatasetLookup(output_dataset_directory, output_dataset_paths)

    return _LocalRunContext(input_datasets, output_datasets, runtime_parameters)


async def run_async(service: typing.Union[services.Service, typing.Callable], load_context: contexts.ServiceContext = None, input_dataset_paths: typing.Dict[str, str] = None, input_dataset_directory: str = None, output_dataset_directory: str = None, output_dataset_paths: typing.Dict[str, str] = None, split_dataset_name: str = None, runtime_parameters: dict = None, assess_accuracy: dict = None, profile_processing_to_file: str = None):
    if input_dataset_paths is None and input_dataset_directory is None:
        logging.warn("Neither input_dataset_paths nor input_dataset_directory was specified, meaning input datasets will not be available!")
    
    if callable(service):
        service = service()
        initialized_service = True
    else:
        initialized_service = False

    load_time = await _call_load(service, load_context)
    
    print("Running...")

    run_context = _get_run_context(input_dataset_paths, input_dataset_directory, output_dataset_directory, output_dataset_paths, runtime_parameters)

    profiler = get_profiler(profile_processing_to_file)

    process_timer = _ServiceProcessTimer(service, profiler)

    if split_dataset_name:
        for ctx in _get_run_contexts(run_context, split_dataset_name):
            await process_timer.process(ctx)
    else:
        await process_timer.process(run_context)

    profiler.save()

    print("Load time: {}s".format(load_time))
    process_timer.print_stats()

    if initialized_service and hasattr(service, 'dispose'):
        service.dispose()
    
    result = dict(run_context.output_dataframes())

    if assess_accuracy is not None:
        await _perform_accuracy_assessment(run_context, assess_accuracy)

    return result

def run(service: typing.Union[services.Service, typing.Callable], **kwargs):

    loop = asyncio.get_event_loop()

    return loop.run_until_complete(run_async(service, **kwargs))




class BadRequestError(RuntimeError):
    def __init__(self, message: str = None):
        super().__init__(message)
        self.message = message or "The request was not valid"

class BadDatasetError(BadRequestError):
    def __init__(self, dataset_name: str, message: str = None):
        super().__init__(message or "The input dataset '{}' was not valid".format(dataset_name))
        self.name = dataset_name

class DatasetFieldError(BadDatasetError):
    def __init__(self, dataset_name: str, field_name: str, message: str = None):
        super().__init__(dataset_name, message or "The field '{}' was not valid for input dataset '{}'".format(field_name, dataset_name))
        self.field_name = field_name

class MissingDatasetFieldError(DatasetFieldError):
    def __init__(self, dataset_name: str, field_name: str, message: str = None):
        super().__init__(dataset_name, field_name, message or "The field '{}' on input dataset '{}' is required but could not be found!".format(field_name, dataset_name))

class MissingDatasetError(BadDatasetError):
    def __init__(self, dataset_name: str, message: str = None):
        super().__init__(dataset_name, message or "The input dataset '{}' was not found.".format(dataset_name))


class BadParameterError(BadRequestError):
    def __init__(self, name: str, message: str = None):
        super().__init__(message or "The request parameter '{}' was not valid".format(name))
        self.name = name

class MissingParameterError(BadParameterError):
    def __init__(self, name: str, message: str = None):
        super().__init__(name, message or "The input parameter '{}' was not found.".format(name))



import os

from setuptools import find_namespace_packages, find_packages, setup

with open(os.path.join(os.path.dirname(__file__), "README.md"), "r") as fh:
    long_description = fh.read()

with open(os.path.join(os.path.dirname(__file__), "requirements.txt"), "r") as fh:
    install_requires = fh.read().split("\n")

setup(
   name='mlservicewrapper-core',
   use_scm_version = True,
   description='Configure a Python service for repeated execution',
   author='Matthew Haugen',
   author_email='mhaugen@haugenapplications.com',

   url="https://github.com/ml-service-wrapper/ml-service-wrapper-core",
   long_description=long_description,
   long_description_content_type="text/markdown",

   package_dir={"": "src"},
   packages=find_namespace_packages("src", include=['mlservicewrapper.*']),

   install_requires=install_requires,
   #namespace_packages=['mlservicewrapper.core.errors'],
   setup_requires=['setuptools_scm'],
   zip_safe=False,
   python_requires='>=3.6'
)
